{-# LANGUAGE MultiParamTypeClasses, FunctionalDependencies #-}
module Graphics.LambdaCube.RenderSystem where

import Data.Word
import Data.Maybe
import qualified Data.Set as Set
import Foreign.C.Types
import Control.Monad

import Graphics.LambdaCube.Types
import Graphics.LambdaCube.Common
import Graphics.LambdaCube.BlendMode
import Graphics.LambdaCube.Texture
import Graphics.LambdaCube.GpuProgram
import Graphics.LambdaCube.TextureUnitState
import Graphics.LambdaCube.Image
import Graphics.LambdaCube.Light
import Graphics.LambdaCube.RenderSystemCapabilities
import Graphics.LambdaCube.RenderOperation
import Graphics.LambdaCube.HardwareBuffer
import Graphics.LambdaCube.HardwareVertexBuffer
import Graphics.LambdaCube.HardwareIndexBuffer
import Graphics.LambdaCube.HardwareOcclusionQuery
import Graphics.LambdaCube.PixelFormat
import Graphics.LambdaCube.Pass

-- | Enum describing the ways to generate texture coordinates
data TexCoordCalcMethod
    = TEXCALC_NONE                       -- ^ No calculated texture coordinates
    | TEXCALC_ENVIRONMENT_MAP            -- ^ Environment map based on vertex normals
    | TEXCALC_ENVIRONMENT_MAP_PLANAR     -- ^ Environment map based on vertex positions
    | TEXCALC_ENVIRONMENT_MAP_REFLECTION -- ^ Environment map based on vertex positions
    | TEXCALC_ENVIRONMENT_MAP_NORMAL     -- ^ Environment map based on vertex positions
    | TEXCALC_PROJECTIVE_TEXTURE         -- ^ Projective texture
    deriving Eq

-- | Enum describing the various actions which can be taken onthe stencil buffer
data StencilOperation
    = SOP_KEEP           -- ^ Leave the stencil buffer unchanged
    | SOP_ZERO           -- ^ Set the stencil value to zero
    | SOP_REPLACE        -- ^ Set the stencil value to the reference value
    | SOP_INCREMENT      -- ^ Increase the stencil value by 1, clamping at the maximum value
    | SOP_DECREMENT      -- ^ Decrease the stencil value by 1, clamping at 0
    | SOP_INCREMENT_WRAP -- ^ Increase the stencil value by 1, wrapping back to 0 when incrementing the maximum value
    | SOP_DECREMENT_WRAP -- ^ Decrease the stencil value by 1, wrapping when decrementing 0
    | SOP_INVERT         -- ^ Invert the bits of the stencil buffer
    deriving Eq

{-| Defines the functionality of a 3D API
@remarks
The RenderSystem class provides a base interface
which abstracts the general functionality of the 3D API
e.g. Direct3D or OpenGL. Whilst a few of the general
methods have implementations, most of this class is
abstract, requiring a subclass based on a specific API
to be constructed to provide the full functionality.
Note there are 2 levels to the interface - one which
will be used often by the caller of the Ogre library,
and one which is at a lower level and will be used by the
other classes provided by Ogre. These lower level
methods are prefixed with '_' to differentiate them.
The advanced user of the library may use these lower
level methods to access the 3D API at a more fundamental
level (dealing direct with render states and rendering
primitives), but still benefiting from Ogre's abstraction
of exactly which 3D API is in use.
@author
Steven Streeting
@version
1.0
-}
class (HardwareVertexBuffer vb, HardwareIndexBuffer ib, HardwareOcclusionQuery q, Texture t, GpuProgram p, LinkedGpuProgram lp) => RenderSystem a vb ib q t p lp | a -> vb ib q t p lp where
    {-| Create a hardware vertex buffer.
    @remarks
        This method creates a new vertex buffer; this will act as a source of geometry
        data for rendering objects. Note that because the meaning of the contents of
        the vertex buffer depends on the usage, this method does not specify a
        vertex format; the user of this buffer can actually insert whatever data 
        they wish, in any format. However, in order to use this with a RenderOperation,
        the data in this vertex buffer will have to be associated with a semantic element
        of the rendering pipeline, e.g. a position, or texture coordinates. This is done 
        using the VertexDeclaration class, which itself contains VertexElement structures
        referring to the source data.
    @remarks Note that because vertex buffers can be shared, they are reference
        counted so you do not need to worry about destroying themm this will be done
        automatically.
    @param vertexSize The size in bytes of each vertex in this buffer; you must calculate
        this based on the kind of data you expect to populate this buffer with.
    @param numVerts The number of vertices in this buffer.
    @param usage One or more members of the HardwareBuffer::Usage enumeration; you are
        strongly advised to use HBU_STATIC_WRITE_ONLY wherever possible, if you need to 
        update regularly, consider HBU_DYNAMIC_WRITE_ONLY and useShadowBuffer=true.
    @param useShadowBuffer If set to true, this buffer will be 'shadowed' by one stored in 
        system memory rather than GPU or AGP memory. You should set this flag if you intend 
        to read data back from the vertex buffer, because reading data from a buffer
    	in the GPU or AGP memory is very expensive, and is in fact impossible if you
        specify HBU_WRITE_ONLY for the main buffer. If you use this option, all 
        reads and writes will be done to the shadow buffer, and the shadow buffer will
        be synchronised with the real buffer at an appropriate time.
    -}
    createVertexBuffer              :: a -> Int -> Int -> Usage -> Bool -> IO vb
    {-| Create a hardware index buffer.
    @remarks Note that because buffers can be shared, they are reference
        counted so you do not need to worry about destroying them this will be done
        automatically.
    @param itype The type in index, either 16- or 32-bit, depending on how many vertices
    	you need to be able to address
    @param numIndexes The number of indexes in the buffer
    @param usage One or more members of the HardwareBuffer::Usage enumeration.
    @param useShadowBuffer If set to true, this buffer will be 'shadowed' by one stored in 
        system memory rather than GPU or AGP memory. You should set this flag if you intend 
        to read data back from the index buffer, because reading data from a buffer
    	in the GPU or AGP memory is very expensive, and is in fact impossible if you
        specify HBU_WRITE_ONLY for the main buffer. If you use this option, all 
        reads and writes will be done to the shadow buffer, and the shadow buffer will
        be synchronised with the real buffer at an appropriate time.
    -}
    createIndexBuffer               :: a -> IndexType -> Int -> Usage -> Bool -> IO ib

    {-
        virtual TexturePtr createManual(const String & name, const String& group,
            TextureType texType, uint width, uint height, uint depth, 
			int num_mips, PixelFormat format, int usage = TU_DEFAULT, ManualResourceLoader* loader = 0,
			bool hwGammaCorrection = false, uint fsaa = 0, const String& fsaaHint = StringUtil::BLANK);
            TextureUsage = bufferUsage::Usage, autoMipmap::Bool, isTarget::Bool
    -}
    createTexture                   :: a -> String -> TextureType -> Int -> Int -> Int -> TextureMipmap -> PixelFormat -> TextureUsage -> Bool -> Int -> String -> Maybe [Image] -> IO t
    --FIXME: TEMP HACK!!!!!
    dirtyHackCopyTexImage           :: a -> t -> Int -> Int -> Int -> Int -> IO ()
    createGpuProgram                :: a -> GpuProgramType -> String -> IO (Either p String) -- TODO
    createLinkedGpuProgram          :: a -> [p] -> IO (Either lp String) -- TODO
    getName                         :: a -> String                                  -- ^ Returns the name of the rendering system.
    createOcclusionQuery            :: a -> IO q                                    -- ^ Create an object for performing hardware occlusion queries. 
--    destroyOcclusionQuery           :: a -> q -> IO ()                              -- ^ Destroy a hardware occlusion query object. 
--obsolete    createRenderSystemCapabilities  :: a -> IO RenderSystemCapabilities             -- ^ Query the real capabilities of the GPU and driver in the RenderSystem
--    reinitialise                    :: a -> IO ()                                   -- ^ Restart the renderer (normally following a change in settings).
--    shutdown                        :: a -> IO ()                                   -- ^ Shutdown the renderer and cleanup resources.
    setAmbientLight                 :: a -> Float -> Float -> Float -> IO ()        -- ^ Sets the colour & strength of the ambient (global directionless) light in the world.
    setShadingType                  :: a -> ShadeOptions -> IO ()                   -- ^ Sets the type of light shading required (default = Gouraud).

    {-| Sets whether or not dynamic lighting is enabled.
    @param
    enabled If true, dynamic lighting is performed on geometry with normals supplied, geometry without
    normals will not be displayed. If false, no lighting is applied and all geometry will be full brightness.
    -}
    setLightingEnabled              :: a -> Bool -> IO ()

    {-| Sets whether or not W-buffers are enabled if they are available for this renderer.
    @param
    enabled If true and the renderer supports them W-buffers will be used.  If false 
    W-buffers will not be used even if available.  W-buffers are enabled by default 
    for 16bit depth buffers and disabled for all other depths.
    -}
    setWBufferEnabled               :: a -> Bool -> IO ()
--    getWBufferEnabled               :: a -> IO Bool                                 -- ^ Returns true if the renderer will try to use W-buffers when avalible.

    {-|	Create a MultiRenderTarget, which is a render target that renders to multiple RenderTextures
    at once. Surfaces can be bound and unbound at will.
    This fails if mCapabilities->getNumMultiRenderTargets() is smaller than 2.
    -}
    -- TODO: virtual MultiRenderTarget * createMultiRenderTarget(const String & name) = 0; 

    {-| Destroys a render texture -}
    -- TODO: virtual void destroyRenderTexture(const String& name);
    {-| Destroys a render target of any sort -}
    -- TODO: virtual void destroyRenderTarget(const String& name);

    {-| Attaches the passed render target to the render system.
    -}
    -- TODO: virtual void attachRenderTarget( RenderTarget &target );
    {-| Returns a pointer to the render target with the passed name, or NULL if that
    render target cannot be found.
    -}
    -- TODO: virtual RenderTarget * getRenderTarget( const String &name );
    {-| Detaches the render target with the passed name from the render system and
    returns a pointer to it.
    @note
    If the render target cannot be found, NULL is returned.
    -}
    -- TODO: virtual RenderTarget * detachRenderTarget( const String &name );

    -- | Iterator over RenderTargets
    -- TODO: typedef MapIterator<Ogre::RenderTargetMap> RenderTargetIterator;

    {-| Returns a specialised MapIterator over all render targets attached to the RenderSystem. -}
    -- TODO: virtual RenderTargetIterator getRenderTargetIterator(void) {
    -- TODO: 	return RenderTargetIterator( mRenderTargets.begin(), mRenderTargets.end() );
    -- TODO: }

    {-| Returns a description of an error code.
    -}
    -- TODO: virtual String getErrorDescription(long errorNumber) const = 0;

    {-| Defines whether or now fullscreen render windows wait for the vertical blank before flipping buffers.
    @remarks
    By default, all rendering windows wait for a vertical blank (when the CRT beam turns off briefly to move
    from the bottom right of the screen back to the top left) before flipping the screen buffers. This ensures
    that the image you see on the screen is steady. However it restricts the frame rate to the refresh rate of
    the monitor, and can slow the frame rate down. You can speed this up by not waiting for the blank, but
    this has the downside of introducing 'tearing' artefacts where part of the previous frame is still displayed
    as the buffers are switched. Speed vs quality, you choose.
    @note
    Has NO effect on windowed mode render targets. Only affects fullscreen mode.
    @param
    enabled If true, the system waits for vertical blanks - quality over speed. If false it doesn't - speed over quality.
    -}
    setWaitForVerticalBlank         :: a -> Bool -> IO ()
--    getWaitForVerticalBlank         :: a -> IO Bool                                 -- ^ Returns true if the system is synchronising frames with the monitor vertical blank.

    -- ------------------------------------------------------------------------
    --                     Internal Rendering Access
    -- All methods below here are normally only called by other OGRE classes
    -- They can be called by library user if required
    -- ------------------------------------------------------------------------


    {-| Tells the rendersystem to use the attached set of lights (and no others) 
    up to the number specified (this allows the same list to be used with different
    count limits) -}
    useLights                      :: a -> [(Matrix4,Light)] -> IO ()
--    useLights                      :: a -> [Light] -> Int -> IO ()
    {-| Are fixed-function lights provided in view space? Affects optimisation. 
    -}
    -- TODO: virtual bool areFixedFunctionLightsInViewSpace() const { return false; }
    setWorldMatrix                 :: a -> Matrix4 -> IO ()                        -- ^ Sets the world transform matrix.
    --_setWorldMatrices               :: a -> [Matrix4] -> Int -> IO ()               -- ^ Sets multiple world matrices (vertex blending).
    setViewMatrix                  :: a -> Matrix4 -> IO ()                        -- ^ Sets the view transform matrix
    setProjectionMatrix            :: a -> Matrix4 -> IO ()                        -- ^ Sets the projection transform matrix
    
    {-| Sets the surface properties to be used for future rendering.
    
    This method sets the the properties of the surfaces of objects
    to be rendered after it. In this context these surface properties
    are the amount of each type of light the object reflects (determining
    it's colour under different types of light), whether it emits light
    itself, and how shiny it is. Textures are not dealt with here,
    see the _setTetxure method for details.
    This method is used by _setMaterial so does not need to be called
    direct if that method is being used.
    
    @param ambient The amount of ambient (sourceless and directionless)
    light an object reflects. Affected by the colour/amount of ambient light in the scene.
    @param diffuse The amount of light from directed sources that is
    reflected (affected by colour/amount of point, directed and spot light sources)
    @param specular The amount of specular light reflected. This is also
    affected by directed light sources but represents the colour at the
    highlights of the object.
    @param emissive The colour of light emitted from the object. Note that
    this will make an object seem brighter and not dependent on lights in
    the scene, but it will not act as a light, so will not illuminate other
    objects. Use a light attached to the same SceneNode as the object for this purpose.
    @param shininess A value which only has an effect on specular highlights (so
    specular must be non-black). The higher this value, the smaller and crisper the
    specular highlights will be, imitating a more highly polished surface.
    This value is not constrained to 0.0-1.0, in fact it is likely to
    be more (10.0 gives a modest sheen to an object).
    @param tracking A bit field that describes which of the ambient, diffuse, specular
    and emissive colours follow the vertex colour of the primitive. When a bit in this field is set
    its ColourValue is ignored. This is a combination of TVC_AMBIENT, TVC_DIFFUSE, TVC_SPECULAR(note that the shininess value is still
    taken from shininess) and TVC_EMISSIVE. TVC_NONE means that there will be no material property
    tracking the vertex colours.
    -}
    setSurfaceParams               :: a -> ColourValue -> ColourValue -> ColourValue -> ColourValue -> FloatType -> TrackVertexColourType -> IO ()

    {-| Sets whether or not rendering points using OT_POINT_LIST will 
    render point sprites (textured quads) or plain points.
    @param enabled True enables point sprites, false returns to normal
    point rendering.
    -}	
    setPointSpritesEnabled         :: a -> Bool -> IO ()

    {-| Sets the size of points and how they are attenuated with distance.
    @remarks
    When performing point rendering or point sprite rendering,
    point size can be attenuated with distance. The equation for
    doing this is attenuation = 1 / (constant + linear * dist + quadratic * d^2) .
    @par
    For example, to disable distance attenuation (constant screensize) 
    you would set constant to 1, and linear and quadratic to 0. A
    standard perspective attenuation would be 0, 1, 0 respectively.
    -}
    setPointParameters              :: a -> FloatType -> Bool -> FloatType -> FloatType -> FloatType -> FloatType -> FloatType -> IO ()


    setActiveTextureUnit            :: a -> Int -> IO ()
    {-|
    Sets the texture to bind to a given texture unit.

    User processes would not normally call this direct unless rendering
    primitives themselves.

    @param unit The index of the texture unit to modify. Multitexturing 
    hardware can support multiple units (see 
    RenderSystemCapabilites::getNumTextureUnits)
    @param enabled Boolean to turn the unit on/off
    @param texPtr Pointer to the texture to use.
    -}
--    setTexture                      :: a -> Int -> Bool -> Maybe t -> IO ()
    setTexture                      :: a -> Maybe t -> IO ()
    {-|
    Sets the texture to bind to a given texture unit.

    User processes would not normally call this direct unless rendering
    primitives themselves.

    @param unit The index of the texture unit to modify. Multitexturing 
    hardware can support multiple units (see 
    RenderSystemCapabilites::getNumTextureUnits)
    @param enabled Boolean to turn the unit on/off
    @param texname The name of the texture to use - this should have
    already been loaded with TextureManager::load.
    -}
    --virtual void _setTexture(size_t unit, bool enabled, const String &texname);

    {-| Binds a texture to a vertex sampler.
    @remarks
    Not all rendersystems support separate vertex samplers. For those that
    do, you can set a texture for them, separate to the regular texture
    samplers, using this method. For those that don't, you should use the
    regular texture samplers which are shared between the vertex and
    fragment units; calling this method will throw an exception.
    @see RenderSystemCapabilites::getVertexTextureUnitsShared
    -}
    setVertexTexture               :: a -> Maybe t -> IO ()

    {-|
    Sets the texture coordinate set to use for a texture unit.

    Meant for use internally - not generally used directly by apps - the Material and TextureUnitState
    classes let you manage textures far more easily.

    @param unit Texture unit as above
    @param index The index of the texture coordinate set to use.
    -}
--    setTextureCoordSet             :: a -> Int -> Int -> IO ()

    {-|
    Sets a method for automatically calculating texture coordinates for a stage.
    Should not be used by apps - for use by Ogre only.
    @param unit Texture unit as above
    @param m Calculation method to use
    @param frustum Optional Frustum param, only used for projective effects
    -}
    setTextureCoordCalculation     :: a -> TexCoordCalcMethod{- -> Frustum-} -> IO ()

    {-| Sets the texture blend modes from a TextureUnitState record.
    Meant for use internally only - apps should use the Material
    and TextureUnitState classes.
    @param unit Texture unit as above
    @param bm Details of the blending mode
    -}
    setTextureBlendMode            :: a -> LayerBlendModeEx -> LayerBlendModeEx -> IO ()

    {-| Sets the filtering options for a given texture unit.
    @param unit The texture unit to set the filtering options for
    @param minFilter The filter used when a texture is reduced in size
    @param magFilter The filter used when a texture is magnified
    @param mipFilter The filter used between mipmap levels, FO_NONE disables mipmapping
    -}
    setTextureUnitFiltering        :: a -> TextureType -> FilterOptions -> FilterOptions -> FilterOptions -> IO ()

    {-| Sets a single filter for a given texture unit.
    @param unit The texture unit to set the filtering options for
    @param ftype The filter type
    @param filter The filter to be used
    -}
--    setTextureUnitFiltering        :: a -> Int -> FilterType -> FilterOptions -> IO ()

    {-| Sets the maximal anisotropy for the specified texture unit.-}
    setTextureLayerAnisotropy      :: a -> TextureType -> Int -> IO ()

    {-| Sets the texture addressing mode for a texture unit.-}
    setTextureAddressingMode       :: a -> TextureType -> UVWAddressingMode -> IO ()

    {-| Sets the texture border colour for a texture unit.-}
    setTextureBorderColour         :: a -> TextureType -> ColourValue -> IO ()

    {-| Sets the mipmap bias value for a given texture unit.
    @remarks
    This allows you to adjust the mipmap calculation up or down for a
    given texture unit. Negative values force a larger mipmap to be used, 
    positive values force a smaller mipmap to be used. Units are in numbers
    of levels, so +1 forces the mipmaps to one smaller level.
    @note Only does something if render system has capability RSC_MIPMAP_LOD_BIAS.
    -}
    setTextureMipmapBias           :: a -> FloatType -> IO ()

    {-| Sets the texture coordinate transformation matrix for a texture unit.
    @param unit Texture unit to affect
    @param xform The 4x4 matrix
    -}
    setTextureMatrix               :: a -> Matrix4 -> IO ()

    {-| Sets the global blending factors for combining subsequent renders with the existing frame contents.
    The result of the blending operation is:</p>
    <p align="center">final = (texture * sourceFactor) + (pixel * destFactor)</p>
    Each of the factors is specified as one of a number of options, as specified in the SceneBlendFactor
    enumerated type.
    By changing the operation you can change addition between the source and destination pixels to a different operator.
    @param sourceFactor The source factor in the above calculation, i.e. multiplied by the texture colour components.
    @param destFactor The destination factor in the above calculation, i.e. multiplied by the pixel colour components.
    @param op The blend operation mode for combining pixels
    -}
    setSceneBlending               :: a -> SceneBlendFactor -> SceneBlendFactor -> SceneBlendOperation -> IO ()

    {-| Sets the global blending factors for combining subsequent renders with the existing frame contents.
    The result of the blending operation is:</p>
    <p align="center">final = (texture * sourceFactor) + (pixel * destFactor)</p>
    Each of the factors is specified as one of a number of options, as specified in the SceneBlendFactor
    enumerated type.
    @param sourceFactor The source factor in the above calculation, i.e. multiplied by the texture colour components.
    @param destFactor The destination factor in the above calculation, i.e. multiplied by the pixel colour components.
    @param sourceFactorAlpha The source factor in the above calculation for the alpha channel, i.e. multiplied by the texture alpha components.
    @param destFactorAlpha The destination factor in the above calculation for the alpha channel, i.e. multiplied by the pixel alpha components.
    @param op The blend operation mode for combining pixels
    @param alphaOp The blend operation mode for combining pixel alpha values
    -}
    setSeparateSceneBlending       :: a -> SceneBlendFactor -> SceneBlendFactor -> SceneBlendFactor -> SceneBlendFactor -> SceneBlendOperation -> SceneBlendOperation -> IO ()

    {-| Sets the global alpha rejection approach for future renders.
    By default images are rendered regardless of texture alpha. This method lets you change that.
    @param func The comparison function which must pass for a pixel to be written.
    @param val The value to compare each pixels alpha value to (0-255)
    @param alphaToCoverage Whether to enable alpha to coverage, if supported
    -}
    setAlphaRejectSettings         :: a -> CompareFunction -> Int -> Bool -> IO ()

    {-| Notify the rendersystem that it should adjust texture projection to be 
    	relative to a different origin.
    -}
    -- TODO: virtual void _setTextureProjectionRelativeTo(bool enabled, const Vector3& pos);
    {-|
    * Signifies the beginning of a frame, i.e. the start of rendering on a single viewport. Will occur
    * several times per complete frame if multiple viewports exist.
    -}
    -- TODO: virtual void _beginFrame(void) = 0;


    {-|
    * Ends rendering of a frame to the current viewport.
    -}
    -- TODO: virtual void _endFrame(void) = 0;
    {-|
    Sets the provided viewport as the active one for future
    rendering operations. This viewport is aware of it's own
    camera and render target. Must be implemented by subclass.

    @param target Pointer to the appropriate viewport.
    -}
    setViewport                     :: a -> Int -> Int -> Int -> Int -> IO ()
    {-| Get the current active viewport for rendering. -}
    -- TODO: virtual Viewport* _getViewport(void);

    {-| Sets the culling mode for the render system based on the 'vertex winding'.
    A typical way for the rendering engine to cull triangles is based on the
    'vertex winding' of triangles. Vertex winding refers to the direction in
    which the vertices are passed or indexed to in the rendering operation as viewed
    from the camera, and will wither be clockwise or anticlockwise (that's 'counterclockwise' for
    you Americans out there ;) The default is CULL_CLOCKWISE i.e. that only triangles whose vertices
    are passed/indexed in anticlockwise order are rendered - this is a common approach and is used in 3D studio models
    for example. You can alter this culling mode if you wish but it is not advised unless you know what you are doing.
    You may wish to use the CULL_NONE option for mesh data that you cull yourself where the vertex
    winding is uncertain.
    -}
    setCullingMode                  :: a -> CullingMode -> IO ()
--    getCullingMode                  :: a -> IO CullingMode

    {-| Sets the mode of operation for depth buffer tests from this point onwards.
    Sometimes you may wish to alter the behaviour of the depth buffer to achieve
    special effects. Because it's unlikely that you'll set these options for an entire frame,
    but rather use them to tweak settings between rendering objects, this is an internal
    method (indicated by the '_' prefix) which will be used by a SceneManager implementation
    rather than directly from the client application.
    If this method is never called the settings are automatically the same as the default parameters.
    @param depthTest If true, the depth buffer is tested for each pixel and the frame buffer is only updated
    if the depth function test succeeds. If false, no test is performed and pixels are always written.
    @param depthWrite If true, the depth buffer is updated with the depth of the new pixel if the depth test succeeds.
    If false, the depth buffer is left unchanged even if a new pixel is written.
    @param depthFunction Sets the function required for the depth test.
    -}
    setDepthBufferParams           :: a -> Bool -> Bool -> CompareFunction -> IO ()

    {-| Sets whether or not the depth buffer check is performed before a pixel write.
    @param enabled If true, the depth buffer is tested for each pixel and the frame buffer is only updated
    if the depth function test succeeds. If false, no test is performed and pixels are always written.
    -}
    setDepthBufferCheckEnabled     :: a -> Bool -> IO ()
    {-| Sets whether or not the depth buffer is updated after a pixel write.
    @param enabled If true, the depth buffer is updated with the depth of the new pixel if the depth test succeeds.
    If false, the depth buffer is left unchanged even if a new pixel is written.
    -}
    setDepthBufferWriteEnabled     :: a -> Bool -> IO ()
    {-| Sets the comparison function for the depth buffer check.
    Advanced use only - allows you to choose the function applied to compare the depth values of
    new and existing pixels in the depth buffer. Only an issue if the deoth buffer check is enabled
    (see _setDepthBufferCheckEnabled)
    @param  func The comparison between the new depth and the existing depth which must return true
    for the new pixel to be written.
    -}
    --setDepthBufferFunction         :: a -> CompareFunction -> IO ()
    setDepthBufferFunction         :: a -> Bool -> CompareFunction -> IO ()
    {-| Sets whether or not colour buffer writing is enabled, and for which channels. 
    @remarks
    For some advanced effects, you may wish to turn off the writing of certain colour
    channels, or even all of the colour channels so that only the depth buffer is updated
    in a rendering pass. However, the chances are that you really want to use this option
    through the Material class.
    @param red, green, blue, alpha Whether writing is enabled for each of the 4 colour channels. -}
    setColourBufferWriteEnabled    :: a -> Bool -> Bool -> Bool -> Bool -> IO ()
    {-| Sets the depth bias, NB you should use the Material version of this. 
    @remarks
    When polygons are coplanar, you can get problems with 'depth fighting' where
    the pixels from the two polys compete for the same screen pixel. This is particularly
    a problem for decals (polys attached to another surface to represent details such as
    bulletholes etc.).
    @par
    A way to combat this problem is to use a depth bias to adjust the depth buffer value
    used for the decal such that it is slightly higher than the true value, ensuring that
    the decal appears on top.
    @note
    The final bias value is a combination of a constant bias and a bias proportional
    to the maximum depth slope of the polygon being rendered. The final bias
    is constantBias + slopeScaleBias * maxslope. Slope scale biasing is
    generally preferable but is not available on older hardware.
    @param constantBias The constant bias value, expressed as a value in 
    homogeneous depth coordinates.
    @param slopeScaleBias The bias value which is factored by the maximum slope
    of the polygon, see the description above. This is not supported by all
    cards.

    -}
    setDepthBias                   :: a -> FloatType -> FloatType -> IO ()
    {-| Sets the fogging mode for future geometry.
    @param mode Set up the mode of fog as described in the FogMode enum, or set to FOG_NONE to turn off.
    @param colour The colour of the fog. Either set this to the same as your viewport background colour,
    or to blend in with a skydome or skybox.
    @param expDensity The density of the fog in FOG_EXP or FOG_EXP2 mode, as a value between 0 and 1. The default is 1. i.e. completely opaque, lower values can mean
    that fog never completely obscures the scene.
    @param linearStart Distance at which linear fog starts to encroach. The distance must be passed
    as a parametric value between 0 and 1, with 0 being the near clipping plane, and 1 being the far clipping plane. Only applicable if mode is FOG_LINEAR.
    @param linearEnd Distance at which linear fog becomes completely opaque.The distance must be passed
    as a parametric value between 0 and 1, with 0 being the near clipping plane, and 1 being the far clipping plane. Only applicable if mode is FOG_LINEAR.
    -}
    setFog                         :: a -> FogMode -> ColourValue -> FloatType -> FloatType -> FloatType -> IO ()

    {-| The RenderSystem will keep a count of tris rendered, this resets the count. -}
    -- TODO: virtual void _beginGeometryCount(void);
    {-| Reports the number of tris rendered since the last _beginGeometryCount call. -}
    -- TODO: virtual unsigned int _getFaceCount(void) const;
    {-| Reports the number of batches rendered since the last _beginGeometryCount call. -}
    -- TODO: virtual unsigned int _getBatchCount(void) const;
    {-| Reports the number of vertices passed to the renderer since the last _beginGeometryCount call. -}
    -- TODO: virtual unsigned int _getVertexCount(void) const;

    {-| Generates a packed data version of the passed in ColourValue suitable for
    use as with this RenderSystem.
    @remarks
    Since different render systems have different colour data formats (eg
    RGBA for GL, ARGB for D3D) this method allows you to use 1 method for all.
    @param colour The colour to convert
    @param pDest Pointer to location to put the result.
    -}
    -- TODO: virtual void convertColourValue(const ColourValue& colour, uint32* pDest);
    {-| Get the native VertexElementType for a compact 32-bit colour value
    for this rendersystem.
    -}
    -- TODO: virtual VertexElementType getColourVertexElementType(void) const = 0;

    {-| Converts a uniform projection matrix to suitable for this render system.
    @remarks
    Because different APIs have different requirements (some incompatible) for the
    projection matrix, this method allows each to implement their own correctly and pass
    back a generic OGRE matrix for storage in the engine.
    -}
    -- TODO: virtual void _convertProjectionMatrix(const Matrix4& matrix,
	-- TODO:     Matrix4& dest, bool forGpuProgram = false) = 0;

    {-| Builds a perspective projection matrix suitable for this render system.
    @remarks
    Because different APIs have different requirements (some incompatible) for the
    projection matrix, this method allows each to implement their own correctly and pass
    back a generic OGRE matrix for storage in the engine.
    -}
    -- TODO: virtual void _makeProjectionMatrix(const Radian& fovy, Real aspect, Real nearPlane, Real farPlane, 
	-- TODO:     Matrix4& dest, bool forGpuProgram = false) = 0;

    {-| Builds a perspective projection matrix for the case when frustum is
    not centered around camera.
    @remarks
    Viewport coordinates are in camera coordinate frame, i.e. camera is 
    at the origin.
    -}
    -- TODO: virtual void _makeProjectionMatrix(Real left, Real right, Real bottom, Real top, 
    -- TODO: 	Real nearPlane, Real farPlane, Matrix4& dest, bool forGpuProgram = false) = 0;
    {-| Builds an orthographic projection matrix suitable for this render system.
    @remarks
    Because different APIs have different requirements (some incompatible) for the
    projection matrix, this method allows each to implement their own correctly and pass
    back a generic OGRE matrix for storage in the engine.
    -}
    -- TODO: virtual void _makeOrthoMatrix(const Radian& fovy, Real aspect, Real nearPlane, Real farPlane, 
    -- TODO: 	Matrix4& dest, bool forGpuProgram = false) = 0;

    {-| Update a perspective projection matrix to use 'oblique depth projection'.
    @remarks
    This method can be used to change the nature of a perspective 
    transform in order to make the near plane not perpendicular to the 
    camera view direction, but to be at some different orientation. 
    This can be useful for performing arbitrary clipping (e.g. to a 
    reflection plane) which could otherwise only be done using user
    clip planes, which are more expensive, and not necessarily supported
    on all cards.
    @param matrix The existing projection matrix. Note that this must be a
    perspective transform (not orthographic), and must not have already
    been altered by this method. The matrix will be altered in-place.
    @param plane The plane which is to be used as the clipping plane. This
    plane must be in CAMERA (view) space.
    @param forGpuProgram Is this for use with a Gpu program or fixed-function
    -}
    -- TODO: virtual void _applyObliqueDepthProjection(Matrix4& matrix, const Plane& plane, 
    -- TODO: 	bool forGpuProgram) = 0;

    {-| Sets how to rasterise triangles, as points, wireframe or solid polys. -}
    setPolygonMode                 :: a -> PolygonMode -> IO ()

    {-| Turns stencil buffer checking on or off. 
    @remarks
    Stencilling (masking off areas of the rendering target based on the stencil 
    buffer) can be turned on or off using this method. By default, stencilling is
    disabled.
    -}
    setStencilCheckEnabled          :: a -> Bool -> IO ()
    {-| Determines if this system supports hardware accelerated stencil buffer. 
    @remarks
    Note that the lack of this function doesn't mean you can't do stencilling, but
    the stencilling operations will be provided in software, which will NOT be
    fast.
    @par
    Generally hardware stencils are only supported in 32-bit colour modes, because
    the stencil buffer shares the memory of the z-buffer, and in most cards the 
    z-buffer has to be the same depth as the colour buffer. This means that in 32-bit
    mode, 24 bits of the z-buffer are depth and 8 bits are stencil. In 16-bit mode there
    is no room for a stencil (although some cards support a 15:1 depth:stencil option,
    this isn't useful for very much) so 8 bits of stencil are provided in software.
    This can mean that if you use stencilling, your applications may be faster in 
    32-but colour than in 16-bit, which may seem odd to some people.
    -}
    {-virtual bool hasHardwareStencil(void) = 0;-}

    {-| This method allows you to set all the stencil buffer parameters in one call.
    @remarks
    The stencil buffer is used to mask out pixels in the render target, allowing
    you to do effects like mirrors, cut-outs, stencil shadows and more. Each of
    your batches of rendering is likely to ignore the stencil buffer, 
    update it with new values, or apply it to mask the output of the render.
    The stencil test is:<PRE>
    (Reference Value & Mask) CompareFunction (Stencil Buffer Value & Mask)</PRE>
    The result of this will cause one of 3 actions depending on whether the test fails,
    succeeds but with the depth buffer check still failing, or succeeds with the
    depth buffer check passing too.
    @par
    Unlike other render states, stencilling is left for the application to turn
    on and off when it requires. This is because you are likely to want to change
    parameters between batches of arbitrary objects and control the ordering yourself.
    In order to batch things this way, you'll want to use OGRE's separate render queue
    groups (see RenderQueue) and register a RenderQueueListener to get notifications
    between batches.
    @par
    There are individual state change methods for each of the parameters set using 
    this method. 
    Note that the default values in this method represent the defaults at system 
    start up too.
    @param func The comparison function applied.
    @param refValue The reference value used in the comparison
    @param mask The bitmask applied to both the stencil value and the reference value 
    before comparison
    @param stencilFailOp The action to perform when the stencil check fails
    @param depthFailOp The action to perform when the stencil check passes, but the
    depth buffer check still fails
    @param passOp The action to take when both the stencil and depth check pass.
    @param twoSidedOperation If set to true, then if you render both back and front faces 
    (you'll have to turn off culling) then these parameters will apply for front faces, 
    and the inverse of them will happen for back faces (keep remains the same).
    -}
    setStencilBufferParams          :: a -> CompareFunction -> Word32 -> Word32 -> StencilOperation -> StencilOperation -> StencilOperation -> Bool -> IO ()



    {-| Sets the current vertex declaration, ie the source of vertex data. -}
--    setVertexDeclaration            :: a -> VertexDeclaration -> IO ()
    {-| Sets the current vertex buffer binding state. -}
--    setVertexBufferBinding          :: a -> VertexBufferBinding vb -> IO ()

    {-| Sets whether or not normals are to be automatically normalised.
    @remarks
    This is useful when, for example, you are scaling SceneNodes such that
    normals may not be unit-length anymore. Note though that this has an
    overhead so should not be turn on unless you really need it.
    @par
    You should not normally call this direct unless you are rendering
    world geometry; set it on the Renderable because otherwise it will be
    overridden by material settings. 
    -}
    setNormaliseNormals             :: a -> Bool -> IO ()

    {-|
    Render something to the active viewport.

    Low-level rendering interface to perform rendering
    operations. Unlikely to be used directly by client
    applications, since the SceneManager and various support
    classes will be responsible for calling this method.
    Can only be called between _beginScene and _endScene

    @param Iteration count
    @param op A rendering operation instance, which contains
    details of the operation to be performed.
    -}
    render                          :: a -> RenderOperation vb ib -> IO ()
    bindGeometry                    :: a -> RenderOperation vb ib -> [TextureUnitState t] -> IO ()
    unbindGeometry                  :: a -> RenderOperation vb ib -> IO ()

    {-| Gets the capabilities of the render system. -}
    getCapabilities                 :: a -> RenderSystemCapabilities

    {-| Binds a given GpuProgram (but not the parameters). 
    @remarks Only one GpuProgram of each type can be bound at once, binding another
    one will simply replace the existing one.
    -}
    bindLinkedGpuProgram            :: a -> lp -> IO ()    
--    bindGpuProgram                  :: a -> GpuProgram -> IO ()

    {-| Bind Gpu program parameters.
    @param gptype The type of program to bind the parameters to
    @param params The parameters to bind
    @param variabilityMask A mask of GpuParamVariability identifying which params need binding
    -}
--    bindGpuProgramParameters        :: a -> GpuProgramType -> GpuProgramParameters -> Word16 -> IO ()

    {-| Only binds Gpu program parameters used for passes that have more than one iteration rendering
    -}
--    bindGpuProgramPassIterationParameters :: a -> GpuProgramType -> IO ()
    {-| Unbinds GpuPrograms of a given GpuProgramType.
    @remarks
    This returns the pipeline to fixed-function processing for this type.
    -}
    unbindLinkedGpuProgram                :: a -> IO ()
--    unbindGpuProgram                :: a -> GpuProgramType -> IO ()

    {-| Returns whether or not a Gpu program of the given type is currently bound. -}
--    isGpuProgramBound               :: a -> GpuProgramType -> IO Bool

    {-| Sets the user clipping region.
    -}
    -- TODO: virtual void setClipPlanes(const PlaneList& clipPlanes);

    {-| Add a user clipping plane. -}
    -- TODO: virtual void addClipPlane (const Plane &p);
    {-| Add a user clipping plane. -}
    -- TODO: virtual void addClipPlane (Real A, Real B, Real C, Real D);

    {-| Clears the user clipping region.
    -}
    -- TODO: virtual void resetClipPlanes();

    {-| Utility method for initialising all render targets attached to this rendering system. -}
    -- TODO: virtual void _initRenderTargets(void);

    {-| Utility method to notify all render targets that a camera has been removed, 
    in case they were referring to it as their viewer. 
    -}
    -- TODO: virtual void _notifyCameraRemoved(const Camera* cam);

    {-| Internal method for updating all render targets attached to this rendering system. -}
    -- TODO: virtual void _updateAllRenderTargets(bool swapBuffers = true);
    {-| Internal method for swapping all the buffers on all render targets,
    if _updateAllRenderTargets was called with a 'false' parameter. -}
    -- TODO: virtual void _swapAllRenderTargetBuffers(bool waitForVsync = true);

    {-| Gets whether or not vertex windings set should be inverted; this can be important
    for rendering reflections. -}
    -- TODO: virtual bool getInvertVertexWinding(void);

    {-| Sets whether or not vertex windings set should be inverted; this can be important
    for rendering reflections. -}
    -- TODO: virtual void setInvertVertexWinding(bool invert);
    {-| Sets the 'scissor region' ie the region of the target in which rendering can take place.
    @remarks
    This method allows you to 'mask off' rendering in all but a given rectangular area
    as identified by the parameters to this method.
    @note
    Not all systems support this method. Check the RenderSystemCapabilities for the
    RSC_SCISSOR_TEST capability to see if it is supported.
    @param enabled True to enable the scissor test, false to disable it.
    @param left, top, right, bottom The location of the corners of the rectangle, expressed in
    <i>pixels</i>.
    -}
    setScissorTest                  :: a -> Bool -> Int -> Int -> Int -> Int -> IO ()

    {-| Clears one or more frame buffers on the active render target. 
    @param buffers Combination of one or more elements of FrameBufferType
    denoting which buffers are to be cleared
    @param colour The colour to clear the colour buffer with, if enabled
    @param depth The value to initialise the depth buffer with, if enabled
    @param stencil The value to initialise the stencil buffer with, if enabled.
    -}
    clearFrameBuffer                :: a -> FrameBufferType -> ColourValue -> FloatType -> Word16 -> IO ()
    {-| Returns the horizontal texel offset value required for mapping 
    texel origins to pixel origins in this rendersystem.
    @remarks
    Since rendersystems sometimes disagree on the origin of a texel, 
    mapping from texels to pixels can sometimes be problematic to 
    implement generically. This method allows you to retrieve the offset
    required to map the origin of a texel to the origin of a pixel in
    the horizontal direction.
    -}
    getHorizontalTexelOffset        :: a -> IO FloatType
    {-| Returns the vertical texel offset value required for mapping 
    texel origins to pixel origins in this rendersystem.
    @remarks
    Since rendersystems sometimes disagree on the origin of a texel, 
    mapping from texels to pixels can sometimes be problematic to 
    implement generically. This method allows you to retrieve the offset
    required to map the origin of a texel to the origin of a pixel in
    the vertical direction.
    -}
    getVerticalTexelOffset          :: a -> IO FloatType

    {-| Gets the minimum (closest) depth value to be used when rendering
    using identity transforms.
    @remarks
    When using identity transforms you can manually set the depth
    of a vertex; however the input values required differ per
    rendersystem. This method lets you retrieve the correct value.
    @see Renderable::getUseIdentityView, Renderable::getUseIdentityProjection
    -}
    getMinimumDepthInputValue       :: a -> FloatType
    {-| Gets the maximum (farthest) depth value to be used when rendering
    using identity transforms.
    @remarks
    When using identity transforms you can manually set the depth
    of a vertex; however the input values required differ per
    rendersystem. This method lets you retrieve the correct value.
    @see Renderable::getUseIdentityView, Renderable::getUseIdentityProjection
    -}
    getMaximumDepthInputValue       :: a -> FloatType
    {-| set the current multi pass count value.  This must be set prior to 
    calling _render() if multiple renderings of the same pass state are 
    required.
    @param count Number of times to render the current state.
    -}
    -- TODO: virtual void setCurrentPassIterationCount(const size_t count) { mCurrentPassIterationCount = count; }

    {-| Tell the render system whether to derive a depth bias on its own based on 
    the values passed to it in setCurrentPassIterationCount.
    The depth bias set will be baseValue + iteration * multiplier
    @param derive True to tell the RS to derive this automatically
    @param baseValue The base value to which the multiplier should be
    added
    @param multiplier The amount of depth bias to apply per iteration
    @param slopeScale The constant slope scale bias for completeness
    -}
    -- TODO: virtual void setDeriveDepthBias(bool derive, float baseValue = 0.0f,
    -- TODO: 	float multiplier = 0.0f, float slopeScale = 0.0f)
    -- TODO: {
    -- TODO: 	mDerivedDepthBias = derive;
    -- TODO: 	mDerivedDepthBiasBase = baseValue;
    -- TODO: 	mDerivedDepthBiasMultiplier = multiplier;
    -- TODO: 	mDerivedDepthBiasSlopeScale = slopeScale;
    -- TODO: }

    {-|
     * Set current render target to target, enabling its device context if needed
     -}
    -- TODO: virtual void _setRenderTarget(RenderTarget *target) = 0;

class (HardwareVertexBuffer vb, HardwareIndexBuffer ib, Texture t, LinkedGpuProgram lp) => Renderable r vb ib t lp | r -> vb ib t lp where
    prepare :: Matrix4 -> r -> [RenderEntity vb ib t lp]

data (HardwareVertexBuffer vb, HardwareIndexBuffer ib, Texture t, LinkedGpuProgram lp) => RenderEntity vb ib t lp
    = RenderEntity
    { reOperation   :: RenderOperation vb ib
    , rePassList    :: [Pass t lp]
    , reMatrix      :: Matrix4
    }

--setPass :: RenderSystem -> Pass -> IO ()
setPass time rs pass = do
    let rsc     = getCapabilities rs
        caps    = rscCapabilities rsc
        Pass
            { --psName                     :: String                      -- ^ optional name for the pass
    
            -- Colour properties, only applicable in fixed-function passes
              psAmbient                     = ambient
            , psDiffuse                     = diffuse
            , psSpecular                    = specular
            , psEmissive                    = emissive
            , psShininess                   = shininess
            , psTracking                    = vertexColourTracking
    
            -- Blending factors
            , psSourceBlendFactor           = sourceBlendFactor
            , psDestBlendFactor             = destBlendFactor
            , psSourceBlendFactorAlpha      = sourceBlendFactorAlpha
            , psDestBlendFactorAlpha        = destBlendFactorAlpha

            , psSeparateBlend               = separateBlend

            -- Blending operations
            , psBlendOperation              = blendOperation
            , psAlphaBlendOperation         = alphaBlendOperation
            , psSeparateBlendOperation      = separateBlendOperation

            -- Depth buffer settings
            , psDepthCheck                  = depthCheck
            , psDepthWrite                  = depthWrite
            , psDepthFunc                   = depthFunc
            , psDepthBiasConstant           = depthBiasConstant
            , psDepthBiasSlopeScale         = depthBiasSlopeScale
--            , psDepthBiasPerIteration    :: FloatType

            -- Colour buffer settings
            , psColourWrite                 = colourWrite
    
            -- Alpha reject settings
            , psAlphaRejectFunc             = alphaRejectFunc
            , psAlphaRejectVal              = alphaRejectVal
            , psAlphaToCoverageEnabled      = alphaToCoverageEnabled
        
--            , psTransparentSorting       :: Bool                        -- ^ Transparent depth sorting
--            , psTransparentSortingForced :: Bool                        -- ^ Transparent depth sorting forced
    
            -- Culling mode
            , psCullMode                    = cullMode
--            , psManualCullMode           :: ManualCullingMode
    
            , psLightingEnabled             = lightingEnabled
--            , psMaxSimultaneousLights    :: Int                         -- ^ Max simultaneous lights
--            , psStartLight               :: Int                         -- ^ Starting light index
--            , psLightsPerIteration       :: Maybe Int                   -- ^ Run this pass once per light? Iterate per how many lights?
--            , psOnlyLightType            :: Maybe LightTypes            -- ^ Should it only be run for a certain light type?

            , psShadeOptions                = shadeOptions
            , psPolygonMode	                = polygonMode

            -- Normalisation
--            , psNormaliseNormals         :: Bool
            , psPolygonModeOverrideable     = polygonModeOverrideable
    
            -- Fog
            , psFogOverride                 = fogOverride
            , psFogMode                     = fogMode
            , psFogColour                   = fogColour
            , psFogStart                    = fogStart
            , psFogEnd                      = fogEnd
            , psFogDensity                  = fogDensity
    
            , psTextureUnitStates           = textureUnitStates

            , psVertexProgramUsage          = vertexProgramUsage
        --    , psShadowCasterVertexProgramUsage  :: GpuProgramUsage    -- ^ Vertex program details
        --    , psShadowReceiverVertexProgramUsage    :: GpuProgramUsage  -- ^ Vertex program details
            , psFragmentProgramUsage        = fragmentProgramUsage
        --    , psShadowReceiverFragmentProgramUsage  :: GpuProgramUsage  -- ^ Fragment program details
            , psGeometryProgramUsage        = geometryProgramUsage
            , psLinkedGpuProgram            = linkedGpuProgram
        --    , psQueuedForDeletion        :: Bool    -- ^ Is this pass queued for deletion?

--            , psPassIterationCount       :: Int                         -- ^ number of pass iterations to perform

            , psPointSize                   = pointSize
            , psPointMinSize                = pointMinSize
            , psPointMaxSize                = pointMaxSize
            , psPointSpritesEnabled         = pointSpritesEnabled
            , psPointAttenuationEnabled     = pointAttenuationEnabled
--            , psPointAttenuationCoeffs   :: FloatType3                  -- ^ constant, linear, quadratic coeffs

            {-
        		// TU Content type lookups
        		typedef vector<unsigned short>::type ContentTypeLookup;
        		mutable ContentTypeLookup mShadowContentTypeLookup;
        		mutable bool mContentTypeLookupBuilt;
            -}  
--            , psLightScissoring          :: Bool                        -- ^ Scissoring for the light?
--            , psLightClipPlanes          :: Bool                        -- ^ User clip planes for light?
--            , psIlluminationStage        :: IlluminationStage           -- ^ Illumination stage?
            } = pass
    {-
	if (!mSuppressRenderStateChanges || evenIfSuppressed)
	{
		if (mIlluminationStage == IRS_RENDER_TO_TEXTURE && shadowDerivation)
		{
			// Derive a special shadow caster pass from this one
			pass = deriveShadowCasterPass(pass);
		}
		else if (mIlluminationStage == IRS_RENDER_RECEIVER_PASS && shadowDerivation)
		{
			pass = deriveShadowReceiverPass(pass);
		}

        // Tell params about current pass
        mAutoParamDataSource->setCurrentPass(pass);

		bool passSurfaceAndLightParams = true;
		bool passFogParams = true;
    -}
    let passSurfaceAndLightParams = True
        passFogParams = True

    case linkedGpuProgram of
        { Nothing   -> unbindLinkedGpuProgram rs
        ; Just lp   -> bindLinkedGpuProgram rs lp
        }

    when passSurfaceAndLightParams $ do
        -- Set surface reflectance properties, only valid if lighting is enabled
        when lightingEnabled $
             setSurfaceParams rs ambient diffuse specular emissive shininess vertexColourTracking
        -- Dynamic lighting enabled?
        setLightingEnabled rs lightingEnabled
        
    when passFogParams $ do
        -- New fog params can either be from scene or from material
        -- TODO: implement override 
        --setFog rs newFogMode newFogColour newFogDensity newFogStart newFogEnd
        setFog rs fogMode fogColour fogDensity fogStart fogEnd
    -- TODO
    {-
		if (passFogParams)
		{
			// New fog params can either be from scene or from material
			FogMode newFogMode;
			ColourValue newFogColour;
			Real newFogStart, newFogEnd, newFogDensity;
			if (pass->getFogOverride())
			{
				// New fog params from material
				newFogMode = pass->getFogMode();
				newFogColour = pass->getFogColour();
				newFogStart = pass->getFogStart();
				newFogEnd = pass->getFogEnd();
				newFogDensity = pass->getFogDensity();
			}
			else
			{
				// New fog params from scene
				newFogMode = mFogMode;
				newFogColour = mFogColour;
				newFogStart = mFogStart;
				newFogEnd = mFogEnd;
				newFogDensity = mFogDensity;
			}

			/* In D3D, it applies to shaders prior
			to version vs_3_0 and ps_3_0. And in OGL, it applies to "ARB_fog_XXX" in
			fragment program, and in other ways, them maybe access by gpu program via
			"state.fog.XXX".
			*/
	        mDestRenderSystem->_setFog(
		        newFogMode, newFogColour, newFogDensity, newFogStart, newFogEnd);
		}

        // Tell params about ORIGINAL fog
		// Need to be able to override fixed function fog, but still have
		// original fog parameters available to a shader than chooses to use
        mAutoParamDataSource->setFog(
            mFogMode, mFogColour, mFogDensity, mFogStart, mFogEnd);
    -}

    -- The rest of the settings are the same no matter whether we use programs or not

    -- Set scene blending
    case separateBlend of
        { True  -> setSeparateSceneBlending rs sourceBlendFactor destBlendFactor
                                               sourceBlendFactorAlpha destBlendFactorAlpha
                                               blendOperation
                                               (if separateBlendOperation then blendOperation else alphaBlendOperation)
        ; False -> case psSeparateBlendOperation pass of
                    { True  -> setSeparateSceneBlending rs sourceBlendFactor destBlendFactor
                                                           sourceBlendFactor destBlendFactor
                                                           blendOperation alphaBlendOperation
                    ; False -> setSceneBlending rs sourceBlendFactor destBlendFactor blendOperation
                    }
        }

    -- Set point parameters
    let (pac,pal,paq) = psPointAttenuationCoeffs pass -- TODO: refactor
    setPointParameters rs pointSize pointAttenuationEnabled pac pal paq pointMinSize pointMaxSize

    when (Set.member RSC_POINT_SPRITES caps) $
        setPointSpritesEnabled rs pointSpritesEnabled
    
    -- Texture unit settings
    -- TODO
    {-
		Pass::ConstTextureUnitStateIterator texIter =  pass->getTextureUnitStateIterator();
		size_t unit = 0;
		// Reset the shadow texture index for each pass
		size_t startLightIndex = pass->getStartLight();
		size_t shadowTexUnitIndex = 0;
		size_t shadowTexIndex = mShadowTextures.size();
		if (mShadowTextureIndexLightList.size() > startLightIndex)
			shadowTexIndex = mShadowTextureIndexLightList[startLightIndex];
		while(texIter.hasMoreElements())
		{
			TextureUnitState* pTex = texIter.getNext();
			if (!pass->getIteratePerLight() && 
				isShadowTechniqueTextureBased() && 
				pTex->getContentType() == TextureUnitState::CONTENT_SHADOW)
			{
				// Need to bind the correct shadow texture, based on the start light
				// Even though the light list can change per object, our restrictions
				// say that when texture shadows are enabled, the lights up to the
				// number of texture shadows will be fixed for all objects
				// to match the shadow textures that have been generated
				// see Listener::sortLightsAffectingFrustum and
				// MovableObject::Listener::objectQueryLights
				// Note that light iteration throws the indexes out so we don't bind here
				// if that's the case, we have to bind when lights are iterated
				// in renderSingleObject

				TexturePtr shadowTex;
				if (shadowTexIndex < mShadowTextures.size())
				{
					shadowTex = getShadowTexture(shadowTexIndex);
					// Hook up projection frustum
					Camera *cam = shadowTex->getBuffer()->getRenderTarget()->getViewport(0)->getCamera();
					// Enable projective texturing if fixed-function, but also need to
					// disable it explicitly for program pipeline.
					pTex->setProjectiveTexturing(!pass->hasVertexProgram(), cam);
					mAutoParamDataSource->setTextureProjector(cam, shadowTexUnitIndex);
				}
				else
				{
					// Use fallback 'null' shadow texture
					// no projection since all uniform colour anyway
					shadowTex = mNullShadowTexture;
					pTex->setProjectiveTexturing(false);
					mAutoParamDataSource->setTextureProjector(0, shadowTexUnitIndex);

				}
				pTex->_setTexturePtr(shadowTex);

				++shadowTexIndex;
				++shadowTexUnitIndex;
			}
			else if (mIlluminationStage == IRS_NONE && pass->hasVertexProgram())
			{
				// Manually set texture projector for shaders if present
				// This won't get set any other way if using manual projection
				TextureUnitState::EffectMap::const_iterator effi = 
					pTex->getEffects().find(TextureUnitState::ET_PROJECTIVE_TEXTURE);
				if (effi != pTex->getEffects().end())
				{
					mAutoParamDataSource->setTextureProjector(effi->second.frustum, unit);
				}
			}
			mDestRenderSystem->_setTextureUnitSettings(unit, *pTex);
			++unit;
		}
		// Disable remaining texture units
		mDestRenderSystem->_disableTextureUnitsFrom(pass->getNumTextureUnitStates());
    -}
    -- TODO
    mapM_ (uncurry $ setTextureUnitSettings time rs) $ zip [0..] textureUnitStates

    -- Disable remaining texture units
    forM_ [(length textureUnitStates)..((rscNumTextureUnits $ getCapabilities rs) - 1)] $ \tu -> do
        -- TODO: dont disable disabled texunits
        setActiveTextureUnit rs tu
        setTexture rs Nothing

    -- Set up non-texture related material settings
    -- Depth buffer settings
    --FIXME: high level haskell gl binding handles depth check activation and depth function in one function
    setDepthBufferFunction rs depthCheck depthFunc
    {- -- HINT: GL has unified function for this
        setDepthBufferFunction rs depthFunc
        setDepthBufferCheckEnabled rs depthCheck
    -}
    setDepthBufferWriteEnabled rs depthWrite
    setDepthBias rs depthBiasConstant depthBiasSlopeScale

    -- Alpha-reject settings
    setAlphaRejectSettings rs alphaRejectFunc alphaRejectVal alphaToCoverageEnabled

    -- Set colour write mode
    -- Right now we only use on/off, not per-channel
    setColourBufferWriteEnabled rs colourWrite colourWrite colourWrite colourWrite
    -- TODO
    {-
		// Culling mode
		if (isShadowTechniqueTextureBased() 
			&& mIlluminationStage == IRS_RENDER_TO_TEXTURE
			&& mShadowCasterRenderBackFaces
			&& pass->getCullingMode() == CULL_CLOCKWISE)
		{
			// render back faces into shadow caster, can help with depth comparison
			mPassCullingMode = CULL_ANTICLOCKWISE;
		}
		else
		{
			mPassCullingMode = pass->getCullingMode();
		}
    -}
    -- TODO: calc cull mode according illumination stage
    setCullingMode rs cullMode
    -- Shading
    setShadingType rs shadeOptions
    
    -- Polygon mode
    unless polygonModeOverrideable $
        setPolygonMode rs polygonMode
    
    -- TODO
    {-
		// set pass number
    	mAutoParamDataSource->setPassNumber( pass->getIndex() );

		// mark global params as dirty
		mGpuParamsDirty |= (uint16)GPV_GLOBAL;
    -}
    return ()

{-
	const TexturePtr& TextureUnitState::_getTexturePtr(size_t frame) const
	{
		if (mContentType == CONTENT_NAMED)
		{
			if (frame < mFrames.size() && !mTextureLoadFailed)
			{
				ensureLoaded(frame);
				return mFramePtrs[frame];
			}
			else
			{
				// Silent fail with empty texture for internal method
				static TexturePtr nullTexPtr;
				return nullTexPtr;
			}
		}
		else
		{
			// Manually bound texture, no name or loading
			assert(frame < mFramePtrs.size());
			return mFramePtrs[frame];
		}
	}
-}

{-| Utility function for setting all the properties of a texture unit at once.
This method is also worth using over the individual texture unit settings because it
only sets those settings which are different from the current settings for this
unit, thus minimising render state changes.
-}
--setTextureUnitSettings :: RenderSystem a vb ib q t p => a -> Int -> TextureUnitState t -> IO ()
setTextureUnitSettings time rs texUnit tl = do
    -- This method is only ever called to set a texture unit to valid details
    -- The method _disableTextureUnit is called to turn a unit off
    let rsc     = getCapabilities rs
        caps    = rscCapabilities rsc
        TextureUnitState 
            { tusAnimDuration           = animDuration
--            , tusCubic                   :: Bool                -- ^ is this a series of 6 2D textures to make up a cube?
    
            , tusTextureType            = texType
--            , tusDesiredFormat           :: PixelFormat
--            , tusTextureSrcMipmaps       :: Int                 -- ^ Request number of mipmaps

--            , tusTextureCoordSetIndex    :: Int
            , tusAddressMode            = uvw
            , tusBorderColour           = borderColour

            , tusColourBlendMode        = colourBlendMode
--            , tusColourBlendFallbackSrc  :: SceneBlendFactor
--            , tusColourBlendFallbackDest :: SceneBlendFactor

            , tusAlphaBlendMode         = alphaBlendMode
        --        mutable bool mTextureLoadFailed;
        --    , tusTextureLoadFailed       :: Bool
--            , tusIsAlpha                 :: Bool
--            , tusHwGamma                 :: Bool

        --        mutable bool mRecalcTexMatrix;
        --        Real mUMod, mVMod;
        --        Real mUScale, mVScale;
        --        Radian mRotate;
        --        mutable Matrix4 mTexModMatrix;

            , tusMinFilter              = minFilter
            , tusMagFilter              = magFilter
            , tusMipFilter              = mipFilter

            , tusMaxAniso               = maxAniso
            , tusMipmapBias             = mipmapBias

        --        bool mIsDefaultAniso;
        --        bool mIsDefaultFiltering;
            , tusBindingType            = bindingType
--            , tusContentType             :: ContentType      -- ^ Content type of texture (normal loaded texture, auto-texture)

--            , tusFrameNames              :: [String]
            , tusFrames                 = frames
--            , tusName                    :: String          -- ^ optional name for the TUS
--            , tusTextureAlias            :: String          -- ^ optional alias for texture frames
            , tusEffects                = effects
            }   = tl
        texl    = fromMaybe (error "fromJust 12") frames -- TODO: const TexturePtr& tex = tl._getTexturePtr();

    -- Activate TextureUnit
    setActiveTextureUnit rs texUnit
    -- Vertex texture binding?
    unless (null texl) $ do
        let tex = case animDuration of
                { Nothing   -> head texl
                ; Just 0    -> head texl
                ; Just d    -> texl !! (floor $ (fromIntegral $ length texl) * (snd $ properFraction $ time / d))
                }
        case (Set.member RSC_VERTEX_TEXTURE_FETCH caps && (not $ rscVertexTextureUnitsShared rsc)) of
            { True  -> case bindingType of
                { BT_VERTEX -> do
                    -- Bind vertex texture
                    setVertexTexture rs $ Just tex
                    -- bind nothing to fragment unit (hardware isn't shared but fragment
                    -- unit can't be using the same index
                    setTexture rs Nothing
                ; _         -> do
                    -- vice versa
                    setVertexTexture rs Nothing
                    setTexture rs $ Just tex
                }
            ; False -> do
                -- Shared vertex / fragment textures or no vertex texture support
                -- Bind texture (may be blank)
                setTexture rs $ Just tex
            }
    
    -- Set texture layer filtering
    setTextureUnitFiltering rs texType minFilter magFilter mipFilter

    -- Set texture layer filtering
    when (Set.member RSC_ANISOTROPY caps) $
        setTextureLayerAnisotropy rs texType maxAniso

    -- Set mipmap biasing
    when (Set.member RSC_MIPMAP_LOD_BIAS caps) $
        setTextureMipmapBias rs mipmapBias

    -- Set blend modes
    -- Check to see if blending is supported
    when (Set.member RSC_BLENDING caps) $ do
        setTextureBlendMode rs colourBlendMode alphaBlendMode
        --HINT: Obsolete below, due to stateful behaviour
        -- Note, colour before alpha is important
        --setTextureBlendMode rs colourBlendMode 
        --setTextureBlendMode rs alphaBlendMode

    -- Texture addressing mode
    setTextureAddressingMode rs texType uvw
    -- Set texture border colour only if required
    when (  amU uvw == TAM_BORDER ||
            amV uvw == TAM_BORDER ||
            amW uvw == TAM_BORDER ) $
        setTextureBorderColour rs texType borderColour

    -- Set texture effects
    -- TODO
--    mapM_ effects
    setTextureCoordCalculation rs TEXCALC_NONE
    forM_ effects $ \ e -> case teType e of
        { ET_ENVIRONMENT_MAP    -> setTextureCoordCalculation rs TEXCALC_ENVIRONMENT_MAP
        ; _ -> return ()
        }
--    effects
{-
        // Set texture effects
        TextureUnitState::EffectMap::iterator effi;
        // Iterate over new effects
        bool anyCalcs = false;
        for (effi = tl.mEffects.begin(); effi != tl.mEffects.end(); ++effi)
        {
            switch (effi->second.type)
            {
            case TextureUnitState::ET_ENVIRONMENT_MAP:
                if (effi->second.subtype == TextureUnitState::ENV_CURVED)
                {
                    _setTextureCoordCalculation(texUnit, TEXCALC_ENVIRONMENT_MAP);
                    anyCalcs = true;
                }
                else if (effi->second.subtype == TextureUnitState::ENV_PLANAR)
                {
                    _setTextureCoordCalculation(texUnit, TEXCALC_ENVIRONMENT_MAP_PLANAR);
                    anyCalcs = true;
                }
                else if (effi->second.subtype == TextureUnitState::ENV_REFLECTION)
                {
                    _setTextureCoordCalculation(texUnit, TEXCALC_ENVIRONMENT_MAP_REFLECTION);
                    anyCalcs = true;
                }
                else if (effi->second.subtype == TextureUnitState::ENV_NORMAL)
                {
                    _setTextureCoordCalculation(texUnit, TEXCALC_ENVIRONMENT_MAP_NORMAL);
                    anyCalcs = true;
                }
                break;
            case TextureUnitState::ET_UVSCROLL:
			case TextureUnitState::ET_USCROLL:
			case TextureUnitState::ET_VSCROLL:
            case TextureUnitState::ET_ROTATE:
            case TextureUnitState::ET_TRANSFORM:
                break;
            case TextureUnitState::ET_PROJECTIVE_TEXTURE:
                _setTextureCoordCalculation(texUnit, TEXCALC_PROJECTIVE_TEXTURE, 
                    effi->second.frustum);
                anyCalcs = true;
                break;
            }
        }
        // Ensure any previous texcoord calc settings are reset if there are now none
        if (!anyCalcs)
        {
            _setTextureCoordCalculation(texUnit, TEXCALC_NONE);
        }

        // Change tetxure matrix 
        _setTextureMatrix(texUnit, tl.getTextureTransform());
    }
-}
    return ()
{-
    const Matrix4& TextureUnitState::getTextureTransform() const
    {
        if (mRecalcTexMatrix)
            recalcTextureMatrix();
        return mTexModMatrix;

    }
-}
calcTextureMatrix tus = do
    return ()
{-
    //-----------------------------------------------------------------------
    void TextureUnitState::recalcTextureMatrix() const
    {
        // Assumption: 2D texture coords
        Matrix4 xform;

        xform = Matrix4::IDENTITY;
        if (mUScale != 1 || mVScale != 1)
        {
            // Offset to center of texture
            xform[0][0] = 1/mUScale;
            xform[1][1] = 1/mVScale;
            // Skip matrix concat since first matrix update
            xform[0][3] = (-0.5 * xform[0][0]) + 0.5;
            xform[1][3] = (-0.5 * xform[1][1]) + 0.5;
        }

        if (mUMod || mVMod)
        {
            Matrix4 xlate = Matrix4::IDENTITY;

            xlate[0][3] = mUMod;
            xlate[1][3] = mVMod;

            xform = xlate * xform;
        }

        if (mRotate != Radian(0))
        {
            Matrix4 rot = Matrix4::IDENTITY;
            Radian theta ( mRotate );
            Real cosTheta = Math::Cos(theta);
            Real sinTheta = Math::Sin(theta);

            rot[0][0] = cosTheta;
            rot[0][1] = -sinTheta;
            rot[1][0] = sinTheta;
            rot[1][1] = cosTheta;
            // Offset center of rotation to center of texture
            rot[0][3] = 0.5 + ( (-0.5 * cosTheta) - (-0.5 * sinTheta) );
            rot[1][3] = 0.5 + ( (-0.5 * sinTheta) + (-0.5 * cosTheta) );

            xform = rot * xform;
        }

        mTexModMatrix = xform;
        mRecalcTexMatrix = false;

    }
-}